import json
import time

import jwt
import pytest

import globus_sdk
from tests.common import register_api_route

OIDC_CONFIG = {
    "issuer": "https://auth.globus.org",
    "authorization_endpoint": "https://auth.globus.org/v2/oauth2/authorize",
    "userinfo_endpoint": "https://auth.globus.org/v2/oauth2/userinfo",
    "token_endpoint": "https://auth.globus.org/v2/oauth2/token",
    "revocation_endpoint": "https://auth.globus.org/v2/oauth2/token/revoke",
    "jwks_uri": "https://auth.globus.org/jwk.json",
    "response_types_supported": ["code", "token", "token id_token", "id_token"],
    "id_token_signing_alg_values_supported": ["RS512"],
    "scopes_supported": ["openid", "email", "profile"],
    "token_endpoint_auth_methods_supported": ["client_secret_basic"],
    "claims_supported": [
        "at_hash",
        "aud",
        "email",
        "exp",
        "name",
        "nonce",
        "preferred_username",
        "iat",
        "iss",
        "sub",
    ],
    "subject_types_supported": ["public"],
}

JWK = {
    "keys": [
        {
            "alg": "RS512",
            "e": "AQAB",
            "kty": "RSA",
            "n": "73l27Yp7WT2c0Ve7EoGJ13AuKzg-GHU7Mpgx0JKa_hO04gAXSVXRadQy7gmdLLtAK8uBVcV0fHGgsBl4J92t-I7hayiJSLbgbX-sZhI_OfegeOLcSNB9poPS9w60XGqR9buYOW2x-KXXitsmyHXNmg_-1u0uqfKHu9pmST8dcjUYXTM5F3oJpQKeJlSH8daMlDks4xb9Y83EEFRv-ppY965-WTm2NW4pwLlbgGTWFvZ6YS6GTb-mfGwGuzStI0lKZ7dOFx9ryYQ4wSoUVHtIrypT-gbuaT90Z2SkwOH-GaEZJkudctBeGpieOsyC7P40UXpwgGNFy3xoWL4vHpnHmQ",  # noqa: E501
            "use": "sig",
        }
    ]
}
JWK_PEM = jwt.algorithms.RSAAlgorithm.from_jwk(json.dumps(JWK["keys"][0]))

TOKEN_PAYLOAD = {
    "access_token": "auth_access_token",
    "scope": "profile email openid",
    "expires_in": 172800,
    "token_type": "Bearer",
    "resource_server": "auth.globus.org",
    "state": "_default",
    "refresh_token": "auth_refresh_token",
    "other_tokens": [
        {
            "access_token": "transfer_access_token",
            "scope": "urn:globus:auth:scope:transfer.api.globus.org:all",
            "expires_in": 172800,
            "token_type": "Bearer",
            "resource_server": "transfer.api.globus.org",
            "state": "_default",
            "refresh_token": "transfer_refresh",
        }
    ],
    # this is a real ID token
    # and since the JWK is real as well, it should decode correctly
    "id_token": "eyJ0eXAiOiJKV1QiLCJhbGciOiJSUzUxMiJ9.eyJzdWIiOiJjOGFhZDQzZS1kMjc0LTExZTUtYmY5OC04YjAyODk2Y2Y3ODIiLCJvcmdhbml6YXRpb24iOiJHbG9idXMiLCJuYW1lIjoiU3RlcGhlbiBSb3NlbiIsInByZWZlcnJlZF91c2VybmFtZSI6InNpcm9zZW4yQGdsb2J1c2lkLm9yZyIsImlkZW50aXR5X3Byb3ZpZGVyIjoiNDExNDM3NDMtZjNjOC00ZDYwLWJiZGItZWVlY2FiYTg1YmQ5IiwiaWRlbnRpdHlfcHJvdmlkZXJfZGlzcGxheV9uYW1lIjoiR2xvYnVzIElEIiwiZW1haWwiOiJzaXJvc2VuQHVjaGljYWdvLmVkdSIsImxhc3RfYXV0aGVudGljYXRpb24iOjE2MjE0ODEwMDYsImlkZW50aXR5X3NldCI6W3sic3ViIjoiYzhhYWQ0M2UtZDI3NC0xMWU1LWJmOTgtOGIwMjg5NmNmNzgyIiwib3JnYW5pemF0aW9uIjoiR2xvYnVzIiwibmFtZSI6IlN0ZXBoZW4gUm9zZW4iLCJ1c2VybmFtZSI6InNpcm9zZW4yQGdsb2J1c2lkLm9yZyIsImlkZW50aXR5X3Byb3ZpZGVyIjoiNDExNDM3NDMtZjNjOC00ZDYwLWJiZGItZWVlY2FiYTg1YmQ5IiwiaWRlbnRpdHlfcHJvdmlkZXJfZGlzcGxheV9uYW1lIjoiR2xvYnVzIElEIiwiZW1haWwiOiJzaXJvc2VuQHVjaGljYWdvLmVkdSIsImxhc3RfYXV0aGVudGljYXRpb24iOjE2MjE0ODEwMDZ9LHsic3ViIjoiYjZlMjI3ZTgtZGI1Mi0xMWU1LWI2ZmYtYzNiMWNjMjU5ZTBkIiwibmFtZSI6bnVsbCwidXNlcm5hbWUiOiJzaXJvc2VuK2JhZGVtYWlsQGdsb2J1cy5vcmciLCJpZGVudGl0eV9wcm92aWRlciI6IjkyN2Q3MjM4LWY5MTctNGViMi05YWNlLWM1MjNmYTliYTM0ZSIsImlkZW50aXR5X3Byb3ZpZGVyX2Rpc3BsYXlfbmFtZSI6Ikdsb2J1cyBTdGFmZiIsImVtYWlsIjoic2lyb3NlbitiYWRlbWFpbEBnbG9idXMub3JnIiwibGFzdF9hdXRoZW50aWNhdGlvbiI6bnVsbH0seyJzdWIiOiJmN2Y4OWQwYS1kYzllLTExZTUtYWRkMC1hM2NiZDFhNTU5YjMiLCJuYW1lIjpudWxsLCJ1c2VybmFtZSI6InNpcm9zZW4rYmFkZW1haWwyQGdsb2J1cy5vcmciLCJpZGVudGl0eV9wcm92aWRlciI6IjkyN2Q3MjM4LWY5MTctNGViMi05YWNlLWM1MjNmYTliYTM0ZSIsImlkZW50aXR5X3Byb3ZpZGVyX2Rpc3BsYXlfbmFtZSI6Ikdsb2J1cyBTdGFmZiIsImVtYWlsIjoic2lyb3NlbitiYWRlbWFpbDJAZ2xvYnVzLm9yZyIsImxhc3RfYXV0aGVudGljYXRpb24iOm51bGx9XSwiaXNzIjoiaHR0cHM6Ly9hdXRoLmdsb2J1cy5vcmciLCJhdWQiOiI3ZmI1OGUwMC04MzlkLTQ0ZTMtODA0Ny0xMGE1MDI2MTJkY2EiLCJleHAiOjE2MjE2NTM4MTEsImlhdCI6MTYyMTQ4MTAxMSwiYXRfaGFzaCI6IjFQdlVhbmNFdUxfc2cxV1BsNWx1TUVGR2tjTDZQaDh1cWdpVUZzejhkZUEifQ.CtfnFtfM32ICo0euHv9GnpVHFL1jWz0NriPTXAv6w08Ylk9JBJtmB3oMKNSO-1TGoWUPFDp9TFFk6N32VyF0hsVDtT5DT3t5oq0qfqbPrZA3R04HARW0xtcK_ejNDHBmj6wysey3EzjT764XTvcGOe63CKQ_RJm97ulVaseIT0Aet7AYo5tQuOiSOQ70xzL7Oax3W6TrWi3FIAA-PIMSrAJKbsG7imGOVkaIObG9a-X5yTOcrB4IG4Wat-pN_QiCiiOw_LDCF-r455PwalmnSGUugMYfsdL2k3UxqwOMLIppHnx5-UVAzj3mygj8eZTp6imjqxNMdakS3vhG8dtxbw",  # noqa: E501
}

# this is the 'exp' value encoded above
FIXED_JWT_EXPIRATION_TIME = 1621653811


@pytest.fixture
def client():
    # this client ID is the audience for the above id_token
    # the client_id must match the audience in order for the decode to work (we pass it
    # as the audience during decoding)
    return globus_sdk.AuthLoginClient(client_id="7fb58e00-839d-44e3-8047-10a502612dca")


@pytest.fixture(autouse=True)
def register_token_response():
    register_api_route(
        "auth", "/v2/oauth2/token", method="POST", body=json.dumps(TOKEN_PAYLOAD)
    )


@pytest.fixture
def token_response(register_token_response, client):
    return client.oauth2_token(
        {
            "grant_type": "authorization_code",
            "code": "foo",
            "redirect_uri": "https://bar.example.org/",
        }
    )


def test_decode_id_token(token_response):
    register_api_route(
        "auth",
        "/.well-known/openid-configuration",
        method="GET",
        body=json.dumps(OIDC_CONFIG),
    )
    register_api_route("auth", "/jwk.json", method="GET", body=json.dumps(JWK))

    decoded = token_response.decode_id_token(jwt_params={"verify_exp": False})
    assert decoded["preferred_username"] == "sirosen2@globusid.org"


def test_decode_id_token_with_saved_oidc_config(token_response):
    register_api_route("auth", "/jwk.json", method="GET", body=json.dumps(JWK))

    decoded = token_response.decode_id_token(
        openid_configuration=OIDC_CONFIG, jwt_params={"verify_exp": False}
    )
    assert decoded["preferred_username"] == "sirosen2@globusid.org"


def test_decode_id_token_with_saved_oidc_config_and_jwk(token_response):
    decoded = token_response.decode_id_token(
        openid_configuration=OIDC_CONFIG,
        jwk=JWK_PEM,
        jwt_params={"verify_exp": False},
    )
    assert decoded["preferred_username"] == "sirosen2@globusid.org"


def test_invalid_decode_id_token_usage(token_response):
    with pytest.raises(globus_sdk.exc.GlobusSDKUsageError):
        token_response.decode_id_token(jwk=JWK_PEM, jwt_params={"verify_exp": False})


def test_decode_id_token_with_leeway(token_response):
    register_api_route(
        "auth",
        "/.well-known/openid-configuration",
        method="GET",
        body=json.dumps(OIDC_CONFIG),
    )
    register_api_route("auth", "/jwk.json", method="GET", body=json.dumps(JWK))

    # do a decode with a leeway parameter set high enough that the ancient
    # expiration time will be tolerated
    expiration_delta = time.time() - FIXED_JWT_EXPIRATION_TIME
    decoded = token_response.decode_id_token(
        jwt_params={"leeway": expiration_delta + 1}
    )
    assert decoded["preferred_username"] == "sirosen2@globusid.org"



from __future__ import annotations

import pytest

from globus_sdk.testing import load_response


@pytest.mark.parametrize(
    "case_name",
    (
        "project_id_str",
        "project_id_uuid",
        "high_assurance",
        "not_high_assurance",
        "required_mfa",
        "authentication_assurance_timeout",
        "display_name",
        "description",
        "domain_constraints_include",
        "empty_domain_constraints_include",
        "no_domain_constraints_include",
        "domain_constraints_exclude",
        "empty_domain_constraints_exclude",
        "no_domain_constraints_exclude",
    ),
)
def test_create_policy(
    service_client,
    case_name: str,
):
    meta = load_response(service_client.create_policy, case=case_name).metadata

    res = service_client.create_policy(**meta["args"])
    for k, v in meta["response"].items():
        assert res["policy"][k] == v



import uuid

import globus_sdk
from globus_sdk.testing import load_response


def test_submit(compute_client_v2: globus_sdk.ComputeClientV2):
    meta = load_response(compute_client_v2.submit).metadata
    ep_id, func_id = uuid.uuid1(), uuid.uuid1()
    submit_doc = {
        "task_group_id": meta["task_group_id"],
        "create_websocket_queue": False,
        "tasks": [
            [func_id, ep_id, "36\n00\ngASVDAAAAAAAAACMBlJvZG5leZSFlC4=\n12 ..."],
            [func_id, ep_id, "36\n00\ngASVCwAAAAAAAACMBUJvYmJ5lIWULg==\n12 ..."],
        ],
    }

    res = compute_client_v2.submit(data=submit_doc)

    assert res.http_status == 200
    assert res.data["task_group_id"] == meta["task_group_id"]
    results = res.data["results"]
    assert results[0]["task_uuid"] == meta["task_id"]
    assert results[1]["task_uuid"] == meta["task_id_2"]



from unittest import mock

import pytest

from globus_sdk.experimental.transfer_v2.transport import (
    TRANSFER_V2_DEFAULT_RETRY_CHECKS,
)
from globus_sdk.transport import (
    RequestCallerInfo,
    RetryCheckCollection,
    RetryCheckRunner,
    RetryConfig,
    RetryContext,
)
from globus_sdk.transport.default_retry_checks import DEFAULT_RETRY_CHECKS


def test_transfer_only_replaces_checks():
    # their length matches, meaning things line up
    assert len(TRANSFER_V2_DEFAULT_RETRY_CHECKS) == len(DEFAULT_RETRY_CHECKS)

    # also confirm that this holds once loaded
    # if the implementation of the RetryCheckCollection becomes sensitive to
    # the contents of these tuples, this could fail
    default_variant = RetryCheckCollection()
    default_variant.register_many_checks(DEFAULT_RETRY_CHECKS)

    transfer_variant = RetryCheckCollection()
    transfer_variant.register_many_checks(TRANSFER_V2_DEFAULT_RETRY_CHECKS)

    assert len(default_variant) == len(transfer_variant)


@pytest.mark.parametrize(
    "body,status_code,expected_should_retry",
    [
        # single ExternalError should not retry
        (
            {
                "errors": [
                    {
                        "status": 502,
                        "code": "ExternalError",
                        "detail": (
                            "The GCP endpoint is not currently connected to Globus"
                        ),
                    }
                ],
                "meta": {
                    "request_id": "rhvcR0aHX",
                },
            },
            502,
            False,
        ),
        # single EndpointError should not retry
        (
            {
                "errors": [
                    {
                        "status": 502,
                        "code": "EndpointError",
                        "detail": (
                            "This GCSv5 is older than version 5.4.62 and does "
                            "not support local user selection"
                        ),
                    }
                ],
                "meta": {
                    "request_id": "istNh0Zpz",
                },
            },
            502,
            False,
        ),
        # other 502 should retry
        (
            {
                "errors": [
                    {
                        "status": 502,
                        "code": "BadGateway",
                        "detail": "Something went wrong with the proxy server",
                    }
                ],
                "meta": {
                    "request_id": "istNh0Zpz",
                },
            },
            502,
            True,
        ),
        # sandwiched ExternalError should not retry
        (
            {
                "errors": [
                    {
                        "status": 500,
                        "code": "InternalError",
                        "detail": "An internal processing error has occurred",
                    },
                    {
                        "status": 502,
                        "code": "ExternalError",
                        "detail": (
                            "The GCP endpoint is not currently connected to Globus"
                        ),
                    },
                    {
                        "status": 502,
                        "code": "BadGateway",
                        "detail": "Something went wrong with the proxy server",
                    },
                ],
                "meta": {
                    "request_id": "istNh0Zpz",
                },
            },
            500,
            False,
        ),
        # multiple 500s should retry
        (
            {
                "errors": [
                    {
                        "status": 500,
                        "code": "InternalError",
                        "detail": "An internal processing error has occurred",
                    },
                    {
                        "status": 502,
                        "code": "BadGateway",
                        "detail": "Something went wrong with the proxy server",
                    },
                ],
                "meta": {
                    "request_id": "istNh0Zpz",
                },
            },
            500,
            True,
        ),
    ],
)
def test_transfer_v2_default_retry_checks(body, status_code, expected_should_retry):
    retry_config = RetryConfig()
    retry_config.checks.register_many_checks(TRANSFER_V2_DEFAULT_RETRY_CHECKS)
    checker = RetryCheckRunner(retry_config.checks)

    dummy_response = mock.Mock()
    dummy_response.json = lambda: body
    dummy_response.status_code = status_code
    caller_info = RequestCallerInfo(retry_config=retry_config)
    ctx = RetryContext(1, caller_info=caller_info, response=dummy_response)

    assert checker.should_retry(ctx) is expected_should_retry


def test_transfer_v2_default_retry_checks_value_error():
    retry_config = RetryConfig()
    retry_config.checks.register_many_checks(TRANSFER_V2_DEFAULT_RETRY_CHECKS)
    checker = RetryCheckRunner(retry_config.checks)

    # simulate the error not having valid json
    def _raise_value_error():
        raise ValueError()

    dummy_response = mock.Mock()
    dummy_response.json = _raise_value_error
    dummy_response.status_code = 502
    caller_info = RequestCallerInfo(retry_config=retry_config)
    ctx = RetryContext(1, caller_info=caller_info, response=dummy_response)

    assert checker.should_retry(ctx) is True



from globus_sdk.testing.models import RegisteredResponse, ResponseSet

from ._common import SUBMISSION_ID, TASK_ID

RESPONSES = ResponseSet(
    metadata={"submission_id": SUBMISSION_ID, "task_id": TASK_ID},
    default=RegisteredResponse(
        service="transfer",
        method="POST",
        path="/v0.10/transfer",
        json={
            "DATA_TYPE": "transfer_result",
            "code": "Accepted",
            "message": (
                "The transfer has been accepted and a task has been created "
                "and queued for execution"
            ),
            "request_id": "7HgMVYazI",
            "resource": "/v0.10/transfer",
            "submission_id": SUBMISSION_ID,
            "task_id": TASK_ID,
            "task_link": {
                "DATA_TYPE": "link",
                "href": f"/v0.10/task/{TASK_ID}?format=json",
                "rel": "related",
                "resource": "task",
                "title": "related task",
            },
        },
    ),
    failure=RegisteredResponse(
        service="transfer",
        method="POST",
        path="/v0.10/transfer",
        json={
            "code": "ClientError.BadRequest.NoTransferItems",
            "message": "A transfer requires at least one item",
            "request_id": "oUAA6Sq2P",
            "resource": "/v0.10/transfer",
        },
        status=400,
        metadata={
            "request_id": "oUAA6Sq2P",
        },
    ),
)



from globus_sdk.testing.models import RegisteredResponse, ResponseSet

metadata = {
    "id": "daa09846-eb92-11e9-b89c-9cb6d0d9fd63",
    "display_name": "example gateway 1",
}

RESPONSES = ResponseSet(
    metadata=metadata,
    default=RegisteredResponse(
        service="gcs",
        method="DELETE",
        path=f"/storage_gateways/{metadata['id']}",
        json={
            "DATA_TYPE": "result#1.0.0",
            "http_response_code": 200,
            "detail": "success",
            "message": "Operation successful",
            "code": "success",
            "data": [{}],
        },
    ),
    permission_denied_error=RegisteredResponse(
        service="gcs",
        method="DELETE",
        path=f"/storage_gateways/{metadata['id']}",
        status=403,
        json={
            "DATA_TYPE": "result#1.0.0",
            "code": "permission_denied",
            "detail": "",
            "http_response_code": 403,
            "message": None,
        },
        metadata={
            "http_status": 403,
            "code": "permission_denied",
            "message": "",
            **metadata,
        },
    ),
)



import urllib.parse

import pytest

from globus_sdk.testing import RegisteredResponse, get_last_request, load_response
from tests.common import GO_EP1_ID


def _mk_item(*, name, typ, size=0):
    return {
        "DATA_TYPE": "file",
        "group": "tutorial",
        "last_modified": "2018-04-04 18:30:26+00:00",
        "link_group": None,
        "link_last_modified": None,
        "link_size": None,
        "link_target": None,
        "link_user": None,
        "name": name,
        "permissions": "0755" if typ == "dir" else "0644",
        "size": 4096 if typ == "dir" else size,
        "type": typ,
        "user": "snork",
    }


def _mk_ls_data():
    return {
        "DATA": [
            _mk_item(name="foo", typ="dir"),
            _mk_item(name="tempdir1", typ="dir"),
            _mk_item(name=".bashrc", typ="file", size=3771),
            _mk_item(name=".profile", typ="file", size=807),
        ]
    }


@pytest.fixture(autouse=True)
def _setup_ls_response():
    load_response(
        RegisteredResponse(
            service="transfer",
            path=f"/v0.10/operation/endpoint/{GO_EP1_ID}/ls",
            json=_mk_ls_data(),
        ),
    )


def test_operation_ls(client):
    ls_path = f"https://transfer.api.globus.org/v0.10/operation/endpoint/{GO_EP1_ID}/ls"

    # load the tutorial endpoint ls doc
    ls_doc = client.operation_ls(GO_EP1_ID)

    # check that the result is an iterable of file and dir dict objects
    for x in ls_doc:
        assert "DATA_TYPE" in x
        assert x["DATA_TYPE"] in ("file", "dir")

    req = get_last_request()
    assert req.url == ls_path


@pytest.mark.parametrize(
    "kwargs, expected_qs",
    [
        # orderby with a single str
        ({"orderby": "name"}, {"orderby": ["name"]}),
        # orderby with a multiple strs
        (
            {"orderby": ["size DESC", "name", "type"]},
            {"orderby": ["size DESC,name,type"]},
        ),
        # orderby + filter
        (
            {"orderby": "name", "filter": "name:~*.png"},
            {"orderby": ["name"], "filter": ["name:~*.png"]},
        ),
        # local_user
        (
            {"local_user": "my-user"},
            {"local_user": ["my-user"]},
        ),
        # limit+offset
        (
            {"limit": 10, "offset": 5},
            {"limit": ["10"], "offset": ["5"]},
        ),
    ],
)
def test_operation_ls_params(client, kwargs, expected_qs):
    client.operation_ls(GO_EP1_ID, **kwargs)
    req = get_last_request()
    parsed_qs = urllib.parse.parse_qs(urllib.parse.urlparse(req.url).query)
    assert parsed_qs == expected_qs



from globus_sdk.testing.models import RegisteredResponse, ResponseSet

from .._common import FUNCTION_CODE, FUNCTION_ID, FUNCTION_NAME

RESPONSES = ResponseSet(
    metadata={
        "function_id": FUNCTION_ID,
        "function_name": FUNCTION_NAME,
        "function_code": FUNCTION_CODE,
    },
    default=RegisteredResponse(
        service="compute",
        path="/v2/functions",
        method="POST",
        json={"function_uuid": FUNCTION_ID},
    ),
)



from ..collection import StaticScopeCollection, _urn_scope


class _GroupsScopes(StaticScopeCollection):
    resource_server = "groups.api.globus.org"

    all = _urn_scope(resource_server, "all")
    view_my_groups_and_memberships = _urn_scope(
        resource_server, "view_my_groups_and_memberships"
    )


class _NexusScopes(StaticScopeCollection):
    resource_server = "nexus.api.globus.org"

    groups = _urn_scope(resource_server, "groups")


GroupsScopes = _GroupsScopes()
NexusScopes = _NexusScopes()



from tests.common import register_api_route_fixture_file


def test_get_group_policies(groups_client):
    register_api_route_fixture_file(
        "groups",
        "/v2/groups/d3974728-6458-11e4-b72d-123139141556/policies",
        "get_group_policies.json",
    )
    resp = groups_client.get_group_policies("d3974728-6458-11e4-b72d-123139141556")
    assert resp.http_status == 200
    assert resp.data == {
        "is_high_assurance": False,
        "authentication_assurance_timeout": 28800,
        "group_visibility": "private",
        "group_members_visibility": "managers",
        "join_requests": False,
        "signup_fields": [],
    }



import globus_sdk
from globus_sdk.globus_app import UserApp

# Tutorial Client ID - <replace this with your own client>
NATIVE_CLIENT_ID = "61338d24-54d5-408f-a10d-66c06b59f6d2"

# Globus Tutorial Collection 1
# https://app.globus.org/file-manager/collections/6c54cade-bde5-45c1-bdea-f4bd71dba2cc
SRC_COLLECTION = "6c54cade-bde5-45c1-bdea-f4bd71dba2cc"
SRC_PATH = "/share/godata/file1.txt"

# Globus Tutorial Collection 2
# https://app.globus.org/file-manager/collections/31ce9ba0-176d-45a5-add3-f37d233ba47d
DST_COLLECTION = "31ce9ba0-176d-45a5-add3-f37d233ba47d"
DST_PATH = "/~/example-transfer-script-destination.txt"


def main():
    with UserApp("my-simple-transfer", client_id=NATIVE_CLIENT_ID) as app:
        with globus_sdk.TransferClient(app=app) as client:
            submit_transfer(client)


def submit_transfer(transfer_client: globus_sdk.TransferClient):
    # Comment out each of these lines if the referenced collection is either
    #   (1) A guest collection or (2) high assurance.
    transfer_client.add_app_data_access_scope(SRC_COLLECTION)
    transfer_client.add_app_data_access_scope(DST_COLLECTION)

    transfer_request = globus_sdk.TransferData(SRC_COLLECTION, DST_COLLECTION)
    transfer_request.add_item(SRC_PATH, DST_PATH)

    task = transfer_client.submit_transfer(transfer_request)
    print(f"Submitted transfer. Task ID: {task['task_id']}.")


if __name__ == "__main__":
    main()



import json
import logging
import os
import uuid
from unittest import mock

import pytest

import globus_sdk
from globus_sdk import GlobusApp, GlobusAppConfig, GlobusSDKUsageError, UserApp
from globus_sdk.authorizers import NullAuthorizer
from globus_sdk.scopes import Scope, TransferScopes
from globus_sdk.testing import RegisteredResponse, get_last_request
from globus_sdk.token_storage import TokenValidationError
from globus_sdk.transport import RequestsTransport


@pytest.fixture
def auth_client():
    return globus_sdk.NativeAppAuthClient(client_id=uuid.uuid1())


@pytest.fixture
def base_client_class():
    class CustomClient(globus_sdk.BaseClient):
        service_name = "transfer"
        scopes = TransferScopes
        default_scope_requirements = [TransferScopes.all]

        def __init__(self, **kwargs) -> None:
            super().__init__(**kwargs)
            self.retry_config.max_retries = 0

    return CustomClient


@pytest.fixture
def base_client(base_client_class):
    return base_client_class()


# not particularly special, just a handy array of codes which should raise
# errors when encountered
ERROR_STATUS_CODES = (400, 404, 405, 409, 500, 503)


def test_cannot_instantiate_plain_base_client():
    # attempting to instantiate a BaseClient errors
    with pytest.raises(GlobusSDKUsageError):
        globus_sdk.BaseClient()


def test_can_instantiate_base_client_with_explicit_url():
    client = globus_sdk.BaseClient(base_url="https://example.org")
    assert client.base_url == "https://example.org"


def test_can_instantiate_with_base_url_class_attribute():
    class MyCoolClient(globus_sdk.BaseClient):
        base_url = "https://example.org/"

    client = MyCoolClient()
    assert client.base_url == "https://example.org/"


def test_base_url_resolution_precedence():
    """
    Base URL can come from one of 3 different places; this test asserts that we maintain
    a consistent precedence between the three
        (init-base_url > class-base_url > class-service_name)
    """

    class BothAttributesClient(globus_sdk.BaseClient):
        base_url = "class-base"
        service_name = "service-name"

    class OnlyServiceClient(globus_sdk.BaseClient):
        service_name = "service-name"

    # All 3 are set
    assert BothAttributesClient(base_url="init-base").base_url == "init-base"
    assert BothAttributesClient().base_url == "class-base"
    assert OnlyServiceClient().base_url == "https://service-name.api.globus.org/"


def test_set_http_timeout(base_client):
    class FooClient(globus_sdk.BaseClient):
        service_name = "foo"

    with mock.patch.dict(os.environ):
        # ensure not set
        os.environ.pop("GLOBUS_SDK_HTTP_TIMEOUT", None)

        client = FooClient()
        assert client.transport.http_timeout == 60.0

        client = FooClient(transport=RequestsTransport(http_timeout=None))
        assert client.transport.http_timeout == 60.0

        client = FooClient(transport=RequestsTransport(http_timeout=-1))
        assert client.transport.http_timeout is None

        os.environ["GLOBUS_SDK_HTTP_TIMEOUT"] = "120"
        client = FooClient()
        assert client.transport.http_timeout == 120.0

        os.environ["GLOBUS_SDK_HTTP_TIMEOUT"] = "-1"
        client = FooClient()
        assert client.transport.http_timeout is None


@pytest.mark.parametrize("mode", ("init", "post_init"))
def test_set_app_name(base_client, base_client_class, mode):
    """
    Sets app name, confirms results
    """
    # set app name
    if mode == "post_init":
        c = base_client
        base_client.app_name = "SDK Test"
    elif mode == "init":
        c = base_client_class(app_name="SDK Test")
    else:
        raise NotImplementedError

    # confirm results
    assert c.app_name == "SDK Test"
    assert c.transport.user_agent == f"{c.transport.BASE_USER_AGENT}/SDK Test"


@pytest.mark.parametrize(
    "method, allows_body",
    [("get", False), ("delete", False), ("post", True), ("put", True), ("patch", True)],
)
def test_http_methods(method, allows_body, base_client):
    """
    BaseClient.{get, delete, post, put, patch} on a path does "the right thing"
    Sends a text body or JSON body as requested
    Raises a GlobusAPIError if the response is not a 200

    NOTE: tests sending request bodies even on GET (which
    *shouldn't* have bodies but *may* have them in reality).
    """
    methodname = method.upper()
    resolved_method = getattr(base_client, method)
    path = "/v0.10/madeuppath/objectname"
    RegisteredResponse(
        service="transfer", path=path, method=methodname, json={"x": "y"}
    ).add()

    # request with no body
    res = resolved_method(path)
    req = get_last_request()

    assert req.method == methodname
    assert req.body is None
    assert "x" in res
    assert res["x"] == "y"

    if allows_body:
        jsonbody = {"foo": "bar"}
        res = resolved_method(path, data=jsonbody)
        req = get_last_request()

        assert req.method == methodname
        assert req.body == json.dumps(jsonbody).encode("utf-8")
        assert "x" in res
        assert res["x"] == "y"

        res = resolved_method(path, data="abc")
        req = get_last_request()

        assert req.method == methodname
        assert req.body == "abc"
        assert "x" in res
        assert res["x"] == "y"

    # send "bad" request
    for status in ERROR_STATUS_CODES:
        RegisteredResponse(
            service="transfer",
            path=path,
            method=methodname,
            json={"x": "y", "code": "ErrorCode", "message": "foo"},
            status=status,
        ).replace()

        with pytest.raises(globus_sdk.GlobusAPIError) as excinfo:
            resolved_method(path)

        assert excinfo.value.http_status == status
        assert excinfo.value.raw_json["x"] == "y"
        assert excinfo.value.code == "ErrorCode"
        assert excinfo.value.message == "foo"


def test_handle_url_unsafe_chars(base_client):
    # make sure this path (escaped) and the request path (unescaped) match
    RegisteredResponse(
        service="transfer", path="/v0.10/foo/foo%20bar", json={"x": "y"}
    ).add()
    res = base_client.get("/v0.10/foo/foo bar")
    assert "x" in res
    assert res["x"] == "y"


def test_access_resource_server_property_via_instance(base_client):
    # get works (and returns accurate info)
    assert base_client.resource_server == TransferScopes.resource_server


def test_access_resource_server_property_via_class(base_client_class):
    # get works (and returns accurate info)
    assert base_client_class.resource_server == TransferScopes.resource_server


def test_app_integration(base_client_class):
    def _reraise_token_error(_: GlobusApp, error: TokenValidationError):
        raise error

    config = GlobusAppConfig(token_validation_error_handler=_reraise_token_error)
    app = UserApp("SDK Test", client_id="client_id", config=config)

    c = base_client_class(app=app)

    # confirm app_name set
    assert c.app_name == "SDK Test"

    # confirm default_required_scopes were automatically added
    assert [str(s) for s in app.scope_requirements[c.resource_server]] == [
        str(TransferScopes.all)
    ]

    # confirm attempt at getting an authorizer from app
    RegisteredResponse(
        service="transfer", path="foo", method="get", json={"x": "y"}
    ).add()
    with pytest.raises(TokenValidationError) as ex:
        c.get("foo")
    assert str(ex.value) == "No token data for transfer.api.globus.org"


def test_app_scopes(base_client_class):
    app = UserApp("SDK Test", client_id="client_id")
    c = base_client_class(app=app, app_scopes=[Scope("foo")])

    # confirm app_scopes were added and default_required_scopes were not
    assert [str(s) for s in app.scope_requirements[c.resource_server]] == ["foo"]


def test_add_app_scope(base_client_class):
    app = UserApp("SDK Test", client_id="client_id")
    c = base_client_class(app=app)

    c.add_app_scope("foo")
    str_list = [str(s) for s in app.scope_requirements[c.resource_server]]
    assert len(str_list) == 2
    assert str(TransferScopes.all) in str_list
    assert "foo" in str_list


def test_add_app_scope_chaining(base_client_class):
    app = UserApp("SDK Test", client_id="client_id")
    c = base_client_class(app=app).add_app_scope("foo").add_app_scope("bar")
    str_list = [str(s) for s in app.scope_requirements[c.resource_server]]
    assert len(str_list) == 3
    assert str(TransferScopes.all) in str_list
    assert "foo" in str_list
    assert "bar" in str_list


def test_app_mutually_exclusive(base_client_class):
    app = UserApp("SDK Test", client_id="client_id")
    expected = "A CustomClient cannot use both an 'app' and an 'authorizer'."

    authorizer = NullAuthorizer()
    with pytest.raises(globus_sdk.exc.GlobusSDKUsageError) as ex:
        base_client_class(app=app, authorizer=authorizer)
    assert str(ex.value) == expected


def test_app_name_override(base_client_class):
    app = UserApp("SDK Test", client_id="client_id")
    c = base_client_class(app=app, app_name="foo")
    assert c.app_name == "foo"


def test_app_scopes_requires_app(base_client_class):
    with pytest.raises(
        globus_sdk.exc.GlobusSDKUsageError,
        match=r"A CustomClient must have an 'app' to use 'app_scopes'\.",
    ):
        base_client_class(app_scopes=[Scope("foo")])


def test_cannot_double_attach_app(base_client_class):
    app = UserApp("SDK Test", client_id="client_id")
    c = base_client_class(app=app)
    with pytest.raises(
        globus_sdk.exc.GlobusSDKUsageError,
        match=r"Cannot attach GlobusApp to CustomClient when one is already attached\.",
    ):
        c.attach_globus_app(app)


def test_cannot_attach_app_after_manually_setting_app_scopes(base_client_class):
    c = base_client_class()
    c.app_scopes = [Scope("foo")]
    app = UserApp("SDK Test", client_id="client_id")
    with pytest.raises(
        globus_sdk.exc.GlobusSDKUsageError,
        match=(
            r"Cannot attach GlobusApp to CustomClient when `app_scopes` is already "
            r"set\."
        ),
    ):
        c.attach_globus_app(app)


def test_cannot_attach_app_when_authorizer_was_provided(base_client_class):
    c = base_client_class(authorizer=NullAuthorizer())
    app = UserApp("SDK Test", client_id="client_id")
    with pytest.raises(
        globus_sdk.exc.GlobusSDKUsageError,
        match=(
            r"Cannot attach GlobusApp to CustomClient when it has an authorizer "
            r"assigned\."
        ),
    ):
        c.attach_globus_app(app)


def test_cannot_attach_app_when_resource_server_is_not_resolvable():
    class CustomClient(globus_sdk.BaseClient):
        service_name = "transfer"
        default_scope_requirements = [TransferScopes.all]

    c = CustomClient()
    app = UserApp("SDK Test", client_id="client_id")
    with pytest.raises(
        globus_sdk.exc.GlobusSDKUsageError,
        match=(
            r"Unable to use an 'app' with a client with no 'resource_server' defined\."
        ),
    ):
        c.attach_globus_app(app)


def test_cannot_attach_app_with_mismatched_environment(base_client_class):
    c = base_client_class(environment="preview")
    app = UserApp("SDK Test", client_id="client_id")
    with pytest.raises(
        globus_sdk.exc.GlobusSDKUsageError,
        match=(
            r"\[Environment Mismatch\] CustomClient's environment \(preview\) does not "
            r"match the GlobusApp's configured environment \(production\)\."
        ),
    ):
        c.attach_globus_app(app)


def test_client_close_implicitly_closes_internal_transport(base_client_class):
    # test the private _close() method directly
    client = base_client_class()
    with mock.patch.object(client.transport, "close") as transport_close:
        client.close()

        transport_close.assert_called_once()


def test_client_close_debug_logs_internal_transport_close(base_client_class, caplog):
    caplog.set_level(logging.DEBUG)

    client = base_client_class()
    client.close()

    assert "closing resource of type RequestsTransport for CustomClient" in caplog.text


def test_client_close_does_not_close_explicitly_passed_transport(base_client_class):
    # test the private _close() method directly
    client = base_client_class(transport=RequestsTransport())
    with mock.patch.object(client.transport, "close") as transport_close:
        client.close()

        transport_close.assert_not_called()


def test_client_context_manager_exit_calls_close(base_client_class):
    with mock.patch.object(globus_sdk.BaseClient, "close") as client_close_method:
        with base_client_class():
            client_close_method.assert_not_called()
        client_close_method.assert_called_once()



import uuid

from globus_sdk.testing.models import RegisteredResponse, ResponseSet

TUNNEL_ID = str(uuid.uuid4())

_initiator_ap = str(uuid.uuid4())
_listener_ap = str(uuid.uuid4())


RESPONSES = ResponseSet(
    default=RegisteredResponse(
        service="transfer",
        method="GET",
        path=f"/v2/tunnels/{TUNNEL_ID}",
        json={
            "data": {
                "attributes": {
                    "created_time": "2025-12-12T21:11:50.525278",
                    "initiator_ip_address": None,
                    "initiator_port": None,
                    "label": "Buzz Tester",
                    "lifetime_mins": 360,
                    "listener_ip_address": None,
                    "listener_port": None,
                    "restartable": False,
                    "state": "AWAITING_LISTENER",
                    "status": "The tunnel is waiting for listening",
                    "submission_id": "292b0054-7084-46eb-83d6-7a6821b1f77e",
                },
                "id": TUNNEL_ID,
                "relationships": {
                    "initiator": {
                        "data": {"id": _initiator_ap, "type": "StreamAccessPoint"}
                    },
                    "listener": {
                        "data": {"id": _listener_ap, "type": "StreamAccessPoint"}
                    },
                    "owner": {
                        "data": {
                            "id": "4d443580-012d-4954-816f-e0592bd356e1",
                            "type": "Identity",
                        }
                    },
                },
                "type": "Tunnel",
            },
            "meta": {"request_id": "M6kFaS949"},
        },
        metadata={
            "tunnel_id": TUNNEL_ID,
            "initiator_ap": _initiator_ap,
            "listener_ap": _listener_ap,
        },
    ),
)



from globus_sdk.testing.models import RegisteredResponse, ResponseSet

from ._common import TASK_ID

RESPONSES = ResponseSet(
    metadata={"task_id": TASK_ID},
    default=RegisteredResponse(
        service="transfer",
        method="GET",
        path=f"/v0.10/endpoint_manager/task/{TASK_ID}/successful_transfers",
        json={
            "DATA_TYPE": "successful_transfers",
            "marker": 0,
            "next_marker": 93979,
            "DATA": [
                {
                    "destination_path": "/path/to/destination",
                    "source_path": "/path/to/source",
                    "DATA_TYPE": "successful_transfer",
                }
            ],
        },
    ),
)



import typing as t

from globus_sdk import SpecificFlowClient
from globus_sdk.testing import load_response


def test_resume_run(specific_flow_client_class: t.Type[SpecificFlowClient]):
    metadata = load_response(SpecificFlowClient.resume_run).metadata

    flow_client = specific_flow_client_class(flow_id=metadata["flow_id"])

    resp = flow_client.resume_run(metadata["run_id"])
    assert resp.http_status == 200



import json

from globus_sdk.testing import get_last_request, load_response


def test_create_group(groups_client):
    meta = load_response(groups_client.create_group).metadata
    res = groups_client.create_group(
        {"name": "Claptrap's Rough Riders", "description": "No stairs allowed."}
    )

    assert res.http_status == 200
    assert "Claptrap" in res["name"]
    assert "No stairs allowed." in res["description"]
    assert res["id"] == meta["group_id"]

    req = get_last_request()
    req_body = json.loads(req.body)
    assert req_body["description"] == "No stairs allowed."


def test_create_group_via_manager(groups_manager, groups_client):
    meta = load_response(groups_client.create_group).metadata
    res = groups_manager.create_group(
        name="Claptrap's Rough Riders", description="No stairs allowed."
    )

    assert res.http_status == 200
    assert "Claptrap" in res["name"]
    assert "No stairs allowed." in res["description"]
    assert res["id"] == meta["group_id"]

    req = get_last_request()
    req_body = json.loads(req.body)
    assert req_body["description"] == "No stairs allowed."



import uuid

from globus_sdk.testing.models import RegisteredResponse, ResponseSet

TUNNEL_ID = str(uuid.uuid4())

_initiator_ap = str(uuid.uuid4())
_listener_ap = str(uuid.uuid4())

_default_display_name = "Test Tunnel"


RESPONSES = ResponseSet(
    default=RegisteredResponse(
        service="transfer",
        method="POST",
        path="/v2/tunnels",
        json={
            "data": {
                "attributes": {
                    "created_time": "2025-12-12T21:49:22.183977",
                    "initiator_ip_address": None,
                    "initiator_port": None,
                    "label": _default_display_name,
                    "lifetime_mins": 10,
                    "listener_ip_address": None,
                    "listener_port": None,
                    "restartable": False,
                    "state": "AWAITING_LISTENER",
                    "status": "The tunnel is waiting for listening.",
                    "submission_id": "6ab42cda-d7a4-11f0-ad34-0affc202d2e9",
                },
                "id": "34d97133-f17e-4f90-ad42-56ff5f3c2550",
                "relationships": {
                    "initiator": {
                        "data": {"id": _initiator_ap, "type": "StreamAccessPoint"}
                    },
                    "listener": {
                        "data": {"id": _listener_ap, "type": "StreamAccessPoint"}
                    },
                    "owner": {
                        "data": {
                            "id": "4d443580-012d-4954-816f-e0592bd356e1",
                            "type": "Identity",
                        }
                    },
                },
                "type": "Tunnel",
            },
            "meta": {"request_id": "e6KkKkNmw"},
        },
        metadata={
            "tunnel_id": TUNNEL_ID,
            "display_name": _default_display_name,
            "initiator_ap": _initiator_ap,
            "listener_ap": _listener_ap,
        },
    ),
)



from globus_sdk.testing.models import RegisteredResponse, ResponseSet

from .._common import ENDPOINT_ID

DEFAULT_RESPONSE_DOC = {
    "endpoint_id": ENDPOINT_ID,
    "task_queue_info": {
        "connection_url": "amqps://user:password@mq.fqdn",
        "exchange": "some_exchange",
        "queue": "some_queue",
    },
    "result_queue_info": {
        "connection_url": "amqps://user:password@mq.fqdn",
        "exchange": "some_exchange",
        "queue": "some_queue",
        "queue_publish_kwargs": {},
    },
}

RESPONSES = ResponseSet(
    default=RegisteredResponse(
        service="compute",
        path="/v3/endpoints",
        method="POST",
        json=DEFAULT_RESPONSE_DOC,
    ),
)



from __future__ import annotations

import dataclasses


@dataclasses.dataclass
class TokenValidationContext:
    """
    A data object of information available to
    :class:`TokenDataValidators <TokenDataValidator>` during validation.

    :ivar str | None prior_identity_id: The identity ID associated with the
        :class:`ValidatingTokenStorage` before the operation began, if there was one.
    :ivar str | None token_data_identity_id: The identity ID extracted from the token
        data being validated.
    """

    prior_identity_id: str | None
    token_data_identity_id: str | None



from xml.etree import ElementTree

import pytest

# tests below will check the exact contents of BaseClient params
# if BaseClient changes, this will need to update -- but that's rare enough to be
# acceptable
BASE_CLIENT_PARAMS = (
    "app",
    "app_scopes",
    "authorizer",
    "app_name",
    "base_url",
    "transport",
    "retry_config",
)


def test_copy_params_requires_an_argument(
    sphinxext, docutils_runner, register_temporary_directive
):

    register_temporary_directive("sdk-copy-params", sphinxext.directives.CopyParams)

    with pytest.raises(Exception, match=r"1 argument\(s\) required, 0 supplied\."):
        docutils_runner.to_etree(".. sdk-copy-params::")


def test_copy_params_finds_base_client_with_or_without_qualified_name(sphinx_runner):
    etree1 = sphinx_runner.to_etree(
        """\
        .. py:function:: build_client(**kwargs)

            .. sdk-sphinx-copy-params:: globus_sdk.BaseClient
        """,
    )
    etree2 = sphinx_runner.to_etree(
        """\
        .. py:function:: build_client(**kwargs)

            .. sdk-sphinx-copy-params:: BaseClient
        """,
    )

    # the source attribute records the temporary filenames used
    # it doesn't matter what we set them to; just normalize them to be the same
    etree1.set("source", "/dev/stdin")
    etree2.set("source", "/dev/stdin")

    # now render and compare equal
    doc1 = ElementTree.tostring(etree1)
    doc2 = ElementTree.tostring(etree2)

    assert doc1 == doc2


def test_copy_params_renders_params_of_base_client(sphinx_runner):
    etree = sphinx_runner.to_etree(
        """\
        .. py:function:: build_client(**kwargs)

            .. sdk-sphinx-copy-params:: globus_sdk.BaseClient
        """,
    )

    assert etree.tag == "document"

    parameters_field = etree.find("./desc/desc_content/field_list/field")
    assert parameters_field is not None
    assert parameters_field.find("field_name").text == "Parameters"

    parameters_list = parameters_field.find("./field_body/bullet_list")
    assert parameters_list is not None
    parameter_names = parameters_list.findall("./list_item/paragraph/literal_strong")

    assert len(parameter_names) == len(BASE_CLIENT_PARAMS)
    param_names = tuple(p.text for p in parameter_names)
    assert param_names == BASE_CLIENT_PARAMS


def test_copy_params_can_render_after_content(sphinx_runner):
    etree = sphinx_runner.to_etree(
        """\
        .. py:function:: build_client(priority: int = 0, **kwargs)

            .. sdk-sphinx-copy-params:: globus_sdk.BaseClient

                :param priority: How cool this client will be
        """,
    )

    assert etree.tag == "document"

    parameters_field = etree.find("./desc/desc_content/field_list/field")
    assert parameters_field is not None
    assert parameters_field.find("field_name").text == "Parameters"

    parameters_list = parameters_field.find("./field_body/bullet_list")
    assert parameters_list is not None
    parameter_names = parameters_list.findall("./list_item/paragraph/literal_strong")

    assert len(parameter_names) == len(BASE_CLIENT_PARAMS) + 1
    param_names = tuple(p.text for p in parameter_names)
    assert param_names == ("priority",) + BASE_CLIENT_PARAMS


def test_copy_params_can_render_before_content(sphinx_runner):
    etree = sphinx_runner.to_etree(
        """\
        .. py:function:: build_client(**kwargs, priority: int = 0)

            .. sdk-sphinx-copy-params:: globus_sdk.BaseClient

                <YIELD>
                :param priority: How cool this client will be
        """,
    )

    assert etree.tag == "document"

    parameters_field = etree.find("./desc/desc_content/field_list/field")
    assert parameters_field is not None
    assert parameters_field.find("field_name").text == "Parameters"

    parameters_list = parameters_field.find("./field_body/bullet_list")
    assert parameters_list is not None
    parameter_names = parameters_list.findall("./list_item/paragraph/literal_strong")

    assert len(parameter_names) == len(BASE_CLIENT_PARAMS) + 1
    param_names = tuple(p.text for p in parameter_names)
    assert param_names == BASE_CLIENT_PARAMS + ("priority",)


def test_copy_params_can_render_in_the_middle_of_content(sphinx_runner):
    etree = sphinx_runner.to_etree(
        """\
        .. py:function:: build_client(awesomeness: float, **kwargs, priority: int = 0)

            .. sdk-sphinx-copy-params:: globus_sdk.BaseClient

                :param awesomeness: The awesomeness quotient
                <YIELD>
                :param priority: How cool this client will be
        """,
    )

    assert etree.tag == "document"

    parameters_field = etree.find("./desc/desc_content/field_list/field")
    assert parameters_field is not None
    assert parameters_field.find("field_name").text == "Parameters"

    parameters_list = parameters_field.find("./field_body/bullet_list")
    assert parameters_list is not None
    parameter_names = parameters_list.findall("./list_item/paragraph/literal_strong")

    assert len(parameter_names) == len(BASE_CLIENT_PARAMS) + 2
    param_names = tuple(p.text for p in parameter_names)
    assert param_names == ("awesomeness",) + BASE_CLIENT_PARAMS + ("priority",)



import time

import pytest

from globus_sdk import exc
from globus_sdk.scopes import Scope, ScopeCycleError, ScopeParseError, ScopeParser


def test_scope_str_and_repr_simple():
    s = Scope("simple")
    assert str(s) == "simple"
    assert repr(s) == "Scope('simple')"


def test_scope_str_and_repr_optional():
    s = Scope("simple", optional=True)
    assert str(s) == "*simple"
    assert repr(s) == "Scope('simple', optional=True)"


def test_scope_str_and_repr_with_dependencies():
    s = Scope("top")
    s = s.with_dependency(Scope("foo"))
    assert str(s) == "top[foo]"
    s = s.with_dependency(Scope("bar"))
    assert str(s) == "top[foo bar]"
    assert repr(s) == "Scope('top', dependencies=(Scope('foo'), Scope('bar')))"


def test_scope_str_nested():
    bottom = Scope("bottom")
    mid = Scope("mid", dependencies=(bottom,))
    top = Scope("top", dependencies=(mid,))
    assert str(bottom) == "bottom"
    assert str(mid) == "mid[bottom]"
    assert str(top) == "top[mid[bottom]]"


def test_scope_with_optional_dependency_stringifies():
    s = Scope("top")
    s = s.with_dependency(Scope("subscope", optional=True))
    assert str(s) == "top[*subscope]"
    subscope_repr = "Scope('subscope', optional=True)"
    assert repr(s) == f"Scope('top', dependencies=({subscope_repr},))"


def test_scope_parsing_allows_empty_string():
    scopes = ScopeParser.parse("")
    assert scopes == []


@pytest.mark.parametrize(
    "scope_string1,scope_string2",
    [
        ("foo ", "foo"),
        (" foo", "foo"),
        ("foo[ bar]", "foo[bar]"),
    ],
)
def test_scope_parsing_ignores_non_semantic_whitespace(scope_string1, scope_string2):
    list1 = ScopeParser.parse(scope_string1)
    list2 = ScopeParser.parse(scope_string2)
    assert len(list1) == len(list2) == 1
    s1, s2 = list1[0], list2[0]
    # Scope.__eq__ is not defined, so equivalence checking is manual (and somewhat error
    # prone) for now
    assert s1.scope_string == s2.scope_string
    assert s1.optional == s2.optional
    for i in range(len(s1.dependencies)):
        assert s1.dependencies[i].scope_string == s2.dependencies[i].scope_string
        assert s1.dependencies[i].optional == s2.dependencies[i].optional


@pytest.mark.parametrize(
    "scopestring",
    [
        # ending in '*'
        "foo*",
        "foo *",
        # '*' followed by '[] '
        "foo*[bar]",
        "foo *[bar]",
        "foo [bar*]",
        "foo * ",
        "* foo",
        # empty brackets
        "foo[]",
        # starting with open bracket
        "[foo]",
        # double brackets
        "foo[[bar]]",
        # unbalanced open brackets
        "foo[",
        "foo[bar",
        # unbalanced close brackets
        "foo]",
        "foo bar]",
        "foo[bar]]",
        "foo[bar] baz]",
        # space before brackets
        "foo [bar]",
        # missing space before next scope string after ']'
        "foo[bar]baz",
    ],
)
def test_scope_parsing_rejects_bad_inputs(scopestring):
    with pytest.raises(ScopeParseError):
        ScopeParser.parse(scopestring)


@pytest.mark.parametrize(
    "scopestring",
    [
        "foo[foo]",
        "foo[*foo]",
        "foo[bar[foo]]",
        "foo[bar[baz[bar]]]",
        "foo[bar[*baz[bar]]]",
        "foo[bar[*baz[*bar]]]",
    ],
)
def test_scope_parsing_catches_and_rejects_cycles(scopestring):
    with pytest.raises(ScopeCycleError):
        ScopeParser.parse(scopestring)


@pytest.mark.flaky
def test_scope_parsing_catches_and_rejects_very_large_cycles_quickly():
    """
    WARNING: this test is hardware speed dependent and could fail on slow systems.

    This test creates a very long cycle and validates that it can be caught in a
    small timeframe of < 100ms.
    Observed times on a test system were <20ms, and in CI were <60ms.

    Although checking the speed in this way is not ideal, we want to avoid high
    time-complexity in the cycle detection. This test offers good protection against any
    major performance regression.
    """
    scope_string = ""
    for i in range(1000):
        scope_string += f"foo{i}["
    scope_string += " foo10"
    for _ in range(1000):
        scope_string += "]"

    t0 = time.time()
    with pytest.raises(ScopeCycleError):
        ScopeParser.parse(scope_string)
    t1 = time.time()
    assert t1 - t0 < 0.1


@pytest.mark.parametrize(
    "scopestring",
    ("foo", "*foo", "foo[bar]", "foo[*bar]", "foo bar", "foo[bar[baz]]"),
)
def test_scope_parsing_accepts_valid_inputs(scopestring):
    # test *only* that parsing does not error and returns a non-empty list of scopes
    scopes = ScopeParser.parse(scopestring)
    assert isinstance(scopes, list)
    assert len(scopes) > 0
    assert isinstance(scopes[0], Scope)


def test_scope_deserialize_simple():
    scope = Scope.parse("foo")
    assert str(scope) == "foo"


def test_scope_deserialize_with_dependencies():
    # oh, while we're here, let's also check that our whitespace insensitivity works
    scope = Scope.parse("foo[ bar   *baz  ]")
    assert str(scope) in ("foo[bar *baz]", "foo[*baz bar]")


def test_scope_deserialize_fails_on_empty():
    with pytest.raises(ValueError):
        Scope.parse("  ")


def test_scope_deserialize_fails_on_multiple_top_level_scopes():
    with pytest.raises(ValueError):
        Scope.parse("foo bar")


@pytest.mark.parametrize("scope_str", ("*foo", "foo[bar]", "foo[", "foo]", "foo bar"))
def test_scope_init_forbids_special_chars(scope_str):
    with pytest.raises(ValueError):
        Scope(scope_str)


@pytest.mark.parametrize(
    "original, reserialized",
    [
        ("foo[bar *bar]", {"foo[bar]"}),
        ("foo[*bar bar]", {"foo[bar]"}),
        ("foo[bar[baz]] bar[*baz]", {"foo[bar[baz]]", "bar[baz]"}),
        ("foo[bar[*baz]] bar[baz]", {"foo[bar[baz]]", "bar[baz]"}),
    ],
)
def test_scope_parsing_normalizes_optionals(original, reserialized):
    assert {str(s) for s in ScopeParser.parse(original)} == reserialized


@pytest.mark.parametrize(
    "scope_str",
    (
        "foo",
        "*foo",
        "foo[bar] baz",
        " foo  ",
        "foo[bar] bar[foo]",
    ),
)
def test_serialize_of_scope_string_is_exact(scope_str):
    assert ScopeParser.serialize(scope_str) == scope_str


def test_serialize_of_scope_object():
    assert ScopeParser.serialize(Scope("scope1")) == "scope1"


@pytest.mark.parametrize(
    "scope_collection",
    (
        ("scope1",),
        ["scope1"],
        {"scope1"},
        (s for s in ["scope1"]),
    ),
)
def test_serialize_of_simple_collection_of_strings(scope_collection):
    assert ScopeParser.serialize(scope_collection) == "scope1"


@pytest.mark.parametrize(
    "scope_collection, expect_str",
    (
        (("scope1", Scope("scope2")), "scope1 scope2"),
        ((Scope("scope1"), Scope("scope2")), "scope1 scope2"),
        ((Scope("scope1"), Scope("scope2"), "scope3"), "scope1 scope2 scope3"),
        (
            (Scope("scope1"), Scope("scope2"), "scope3 scope4"),
            "scope1 scope2 scope3 scope4",
        ),
        ([Scope("scope1"), "scope2", "scope3 scope4"], "scope1 scope2 scope3 scope4"),
    ),
)
def test_serialize_handles_mixed_data(scope_collection, expect_str):
    assert ScopeParser.serialize(scope_collection) == expect_str


@pytest.mark.parametrize("input_obj", ("", [], set(), ()))
def test_serialize_rejects_empty_by_default(input_obj):
    with pytest.raises(
        exc.GlobusSDKUsageError,
        match="'scopes' cannot be the empty string or empty collection",
    ):
        ScopeParser.serialize(input_obj)


@pytest.mark.parametrize("input_obj", ("", [], set(), ()))
def test_serialize_allows_empty_string_with_flag(input_obj):
    assert ScopeParser.serialize(input_obj, reject_empty=False) == ""



from globus_sdk._internal.classprop import classproperty


def test_classproperty_simple():
    class Foo:
        x = {"x": 1}

        @classproperty
        def y(self_or_cls):
            return self_or_cls.x["x"]

    assert Foo.y == 1


def test_classproperty_prefers_instance():
    class Foo:
        x = {"x": 1}

        def __init__(self) -> None:
            self.x = {"x": 2}

        @classproperty
        def y(self_or_cls):
            return self_or_cls.x["x"]

    assert Foo.y == 1
    assert Foo().y == 2



from globus_sdk.response import IterableResponse


class GetProjectsResponse(IterableResponse):
    """
    Response class specific to the Get Projects API

    Provides iteration on the "projects" array in the response.
    """

    default_iter_key = "projects"



import json

from globus_sdk.experimental import BookmarkCreateDocument
from globus_sdk.testing import get_last_request, load_response


def test_create_bookmark(client):
    meta = load_response(client.create_bookmark).metadata

    data = BookmarkCreateDocument(
        meta["collection"],
        meta["name"],
        meta["path"],
    )

    res = client.create_bookmark(data)

    assert res.http_status == 201
    assert res["data"]["type"] == "Bookmark"

    req = get_last_request()
    sent = json.loads(req.body)
    assert sent["data"]["attributes"]["name"] == meta["name"]
    assert sent["data"]["attributes"]["path"] == meta["path"]
    assert (
        sent["data"]["relationships"]["collection"]["data"]["id"] == meta["collection"]
    )



import pytest

from globus_sdk import MISSING
from globus_sdk.testing import get_last_request, load_response


@pytest.mark.parametrize("include_flow_description", (MISSING, False, True))
def test_get_run(flows_client, include_flow_description):
    metadata = load_response(flows_client.get_run).metadata

    response = flows_client.get_run(
        metadata["run_id"],
        include_flow_description=include_flow_description,
    )
    assert response.http_status == 200

    request = get_last_request()
    if include_flow_description is MISSING:
        assert "flow_description" not in response
        assert "include_flow_description" not in request.url
    elif include_flow_description is False:
        assert "flow_description" not in response
        assert "include_flow_description=False" in request.url
    else:  # include_flow_description is True
        assert "flow_description" in response
        assert "include_flow_description=True" in request.url



#!/usr/bin/env python
from __future__ import annotations

import functools
import glob
import os
import pathlib
import re
import sys
import typing as t

REPO_ROOT = pathlib.Path(__file__).parent.parent

_ALL_NAME_PATTERN = re.compile(r'\s+"(\w+)",?')

PACKAGE_LOCS_TO_SCAN = (
    "globus_sdk/__init__.pyi",
    "globus_sdk/login_flows/",
    "globus_sdk/gare/",
    "globus_sdk/globus_app/",
    "globus_sdk/scopes/",
    "globus_sdk/response.py",
    "globus_sdk/testing/",
)

DEPRECATED_NAMES: set[str] = set()


def load_docs() -> dict[str, str]:
    all_docs = {}
    for file in glob.glob("docs/**/*.rst", recursive=True):
        with open(file) as f:
            all_docs[file] = f.read()
    return all_docs


def iter_all_documented_names() -> t.Iterable[str]:
    # names under these directives
    #
    #   .. autoclass:: <name>
    #   .. autoclass:: <name>(<args>)
    #   .. autofunction:: <name>
    #   .. autoexception:: <name>
    #   .. autodata:: <name>
    autodoc_pattern = re.compile(
        r"""
        ^\.\.\s+auto(?:class|function|exception|data)::\s+ # auto-directive (uncaptured)
        (?:\w+\.)*(\w+)(?:\(.*\))?$                        # symbol name (captured)
        """,
        flags=re.MULTILINE | re.X,
    )
    # names under these directives
    #
    #   .. py:class:: <name>
    #   .. py:data:: <name>
    pydoc_pattern = re.compile(
        r"""
        ^\.\.\s+py:(?:data|class)::\s+ # directive
        (?:\w+\.)*(\w+)(?:\(.*\))?$    # symbol name (captured)
        """,
        flags=re.MULTILINE | re.X,
    )
    for data in load_docs().values():
        for match in autodoc_pattern.finditer(data):
            yield match.group(1)
        for match in pydoc_pattern.finditer(data):
            yield match.group(1)


@functools.lru_cache
def all_documented_names() -> frozenset[str]:
    return frozenset(iter_all_documented_names())


def is_documented(name: str) -> bool:
    return name in all_documented_names()


def get_names_from_all_list(file_path: str) -> list[str]:
    with open(f"src/{file_path}") as fp:
        contents = fp.readlines()

    names: list[str] = []
    found_all = False
    for line in contents:
        if found_all:
            if line.strip() == ")":
                break
            # Extract the actual symbol from the line.
            # i.e., '  "Foo",\n' -> 'Foo'
            name_match = _ALL_NAME_PATTERN.match(line)
            if name_match is not None:
                names.append(name_match.group(1))
        else:
            if line.strip() == "__all__ = (":
                found_all = True
            else:
                continue
    return [n for n in names if not n.startswith("_")]


def ensure_exports_are_documented() -> bool:
    success = True
    used_deprecations = set()
    for loc in PACKAGE_LOCS_TO_SCAN:
        if loc.endswith("/"):
            loc = f"{loc}/__init__.py"
        for name in get_names_from_all_list(loc):
            if name in DEPRECATED_NAMES:
                used_deprecations.add(name)
                continue
            if not is_documented(name):
                print(f"'src/{loc}::{name}' was not found in doc directives")
                success = False
    if unused_deprecations := (DEPRECATED_NAMES - used_deprecations):
        print(f"unused deprecations: {unused_deprecations}")
        success = False
    return success


if __name__ == "__main__":
    os.chdir(REPO_ROOT)
    if not ensure_exports_are_documented():
        exit(1)
    sys.exit(0)



from globus_sdk.testing.models import RegisteredResponse, ResponseSet

from ._common import (
    BASE_GROUP_DOC,
    GROUP_ID,
    SUBSCRIPTION_GROUP_DOC,
    SUBSCRIPTION_GROUP_ID,
    SUBSCRIPTION_ID,
)

RESPONSES = ResponseSet(
    metadata={"group_id": GROUP_ID},
    default=RegisteredResponse(
        service="groups",
        path=f"/v2/groups/{GROUP_ID}",
        json=BASE_GROUP_DOC,
    ),
    subscription=RegisteredResponse(
        service="groups",
        path=f"/v2/groups/{SUBSCRIPTION_GROUP_ID}",
        json=SUBSCRIPTION_GROUP_DOC,
        metadata={
            "group_id": SUBSCRIPTION_GROUP_ID,
            "subscription_id": SUBSCRIPTION_ID,
        },
    ),
)



import uuid

from globus_sdk.testing.models import RegisteredResponse, ResponseSet

SCOPE = {
    "scope_string": "https://auth.globus.org/scopes/3f33d83f-ec0a-4190-887d-0622e7c4ee9a/manager",  # noqa: E501
    "allows_refresh_token": False,
    "id": str(uuid.uuid1()),
    "advertised": False,
    "required_domains": [],
    "name": "Client manage scope",
    "description": "Manage configuration of this client",
    "client": "3f33d83f-ec0a-4190-887d-0622e7c4ee9a",
    "dependent_scopes": [],
}

RESPONSES = ResponseSet(
    default=RegisteredResponse(
        service="auth",
        path=f"/v2/api/scopes/{SCOPE['id']}",
        json={"scope": SCOPE},
        metadata={
            "scope_id": SCOPE["id"],
        },
    ),
)



#!/usr/bin/env python
"""
Check if the version number in the source is already present as a changelog header.
"""

import os
import re
import sys

if sys.version_info >= (3, 11):
    import tomllib
else:
    # Older Python versions
    import tomli as tomllib

PATTERN_FORMAT = "^v{version}\\s+\\({date}\\)$"
CHANGELOG_D = os.path.dirname(__file__)
REPO_ROOT = os.path.dirname(CHANGELOG_D)


def parse_version():
    with open(os.path.join(REPO_ROOT, "pyproject.toml"), "rb") as f:
        pyproject = tomllib.load(f)

    return pyproject["project"]["version"]


def get_header_re(version):
    version = re.escape(version)
    return PATTERN_FORMAT.format(version=version, date=r"\d{4}-\d{2}-\d{2}")


def changelog_has_version(version):
    pattern = get_header_re(version)
    with open(os.path.join(REPO_ROOT, "changelog.rst")) as f:
        return bool(re.search(pattern, f.read(), flags=re.MULTILINE))


def main():
    version = parse_version()
    if changelog_has_version(version):
        print(f"version {version} appears in the changelog already!")
        sys.exit(1)
    else:
        print(f"version check OK ({version} is a new version number)")
    sys.exit(0)


if __name__ == "__main__":
    main()



from __future__ import annotations

import typing as t
import uuid

from globus_sdk._missing import MISSING, MissingType
from globus_sdk._payload import GlobusPayload


class UserCredentialDocument(GlobusPayload):
    """
    Convenience class for constructing a UserCredential document
    to use as the `data` parameter to `create_user_credential` and
    `update_user_credential`

    :param DATA_TYPE: Versioned document type.
    :param identity_id: UUID of the Globus identity this credential will
        provide access for
    :param connector_id: UUID of the connector this credential is for
    :param username: Username of the local account this credential will provide
        access to, format is connector specific
    :param display_name: Display name for this credential
    :param storage_gateway_id: UUID of the storage gateway this credential is for
    :param policies: Connector specific policies for this credential
    :param additional_fields: Additional data for inclusion in the document
    """

    def __init__(
        self,
        DATA_TYPE: str = "user_credential#1.0.0",
        identity_id: uuid.UUID | str | MissingType = MISSING,
        connector_id: uuid.UUID | str | MissingType = MISSING,
        username: str | MissingType = MISSING,
        display_name: str | MissingType = MISSING,
        storage_gateway_id: uuid.UUID | str | MissingType = MISSING,
        policies: dict[str, t.Any] | MissingType = MISSING,
        additional_fields: dict[str, t.Any] | None = None,
    ) -> None:
        super().__init__()
        self["DATA_TYPE"] = DATA_TYPE
        self["identity_id"] = identity_id
        self["connector_id"] = connector_id
        self["username"] = username
        self["display_name"] = display_name
        self["storage_gateway_id"] = storage_gateway_id
        self["policies"] = policies
        self.update(additional_fields or {})



import urllib.parse

import pytest

from globus_sdk.testing import get_last_request, load_response


def test_oauth2_get_dependent_tokens(auth_client):
    meta = load_response(
        auth_client.oauth2_get_dependent_tokens, case="groups"
    ).metadata

    response = auth_client.oauth2_get_dependent_tokens("dummy_token")
    full_response_by_rs = response.by_resource_server

    # data matches fully in the by_resource_server layout
    expected_data = meta["rs_data"]
    assert set(expected_data) == set(full_response_by_rs)
    for rs_name, values in expected_data.items():
        assert full_response_by_rs[rs_name]["access_token"] == values["access_token"]
        assert full_response_by_rs[rs_name]["scope"] == values["scope"]


def test_oauth2_get_dependent_tokens_with_refresh_token(auth_client):
    meta = load_response(
        auth_client.oauth2_get_dependent_tokens, case="groups_with_refresh_token"
    ).metadata

    response = auth_client.oauth2_get_dependent_tokens(
        "dummy_token", refresh_tokens=True
    )
    full_response_by_rs = response.by_resource_server

    # data matches fully in the by_resource_server layout
    expected_data = meta["rs_data"]
    assert set(expected_data) == set(full_response_by_rs)
    for rs_name, values in expected_data.items():
        assert full_response_by_rs[rs_name]["access_token"] == values["access_token"]
        assert full_response_by_rs[rs_name]["refresh_token"] == values["refresh_token"]
        assert full_response_by_rs[rs_name]["scope"] == values["scope"]

    # parse sent request and ensure that refresh_tokens translated correctly
    # to `access_type=offline`
    last_req = get_last_request()
    assert last_req.body
    body = last_req.body
    assert body != ""
    parsed_body = urllib.parse.parse_qs(body)
    assert parsed_body["access_type"] == ["offline"]


@pytest.mark.parametrize(
    "scope_arg, expect_value",
    [(None, None), ("scope1", "scope1"), (("scope1", "scope2"), "scope1 scope2")],
)
def test_oauth2_get_dependent_tokens_scope_string_param(
    auth_client, scope_arg, expect_value
):
    load_response(auth_client.oauth2_get_dependent_tokens, case="groups")

    add_args = {}
    if scope_arg is not None:
        add_args["scope"] = scope_arg
    auth_client.oauth2_get_dependent_tokens("dummy_token", **add_args)

    last_req = get_last_request()
    assert last_req.body
    body = last_req.body
    assert body != ""
    parsed_body = urllib.parse.parse_qs(body)
    if expect_value is None:
        assert "scope" not in parsed_body
    else:
        assert parsed_body["scope"] == [expect_value]



from __future__ import annotations

# the name "datetime" is used in this module, so use an alternative name
# in order to avoid name shadowing
import datetime as dt
import logging
import typing as t
import uuid

from globus_sdk._missing import MISSING, MissingType
from globus_sdk._payload import GlobusPayload
from globus_sdk.services.transfer import TransferData

log = logging.getLogger(__name__)


class TransferTimer(GlobusPayload):
    """
    A helper for defining a payload for Transfer Timer creation.
    Use this along with :meth:`create_timer <globus_sdk.TimersClient.create_timer>` to
    create a timer.

    .. note::
        ``TimersClient`` has two methods for creating timers, ``create_timer`` and
        ``create_job``.
        ``create_job`` uses a different API -- only ``create_timer`` will work with
        this helper class.

        Users are strongly recommended to use ``create_timer`` and this helper for
        timer creation.

    :param name: A name to identify this timer
    :param schedule: The schedule on which the timer runs
    :param body: A transfer payload for the timer to use. If it includes
        ``submission_id`` or ``skip_activation_check``, these parameters will be
        removed, as they are not supported in timers.

    The ``schedule`` field determines when the timer will run.
    Timers may be "run once" or "recurring", and "recurring" timers may specify an end
    date or a number of executions after which the timer will stop. A ``schedule`` is
    specified as a dict, but the SDK provides two useful helpers for constructing these
    data.

    **Example Schedules**

    .. tab-set::

        .. tab-item:: Run Once, Right Now

            .. code-block:: python

                schedule = OnceTimerSchedule()

        .. tab-item:: Run Once, At a Specific Time

            .. code-block:: python

                schedule = OnceTimerSchedule(datetime="2023-09-22T00:00:00Z")

        .. tab-item:: Run Every 5 Minutes, Until a Specific Time

            .. code-block:: python

                schedule = RecurringTimerSchedule(
                    interval_seconds=300,
                    end={"condition": "time", "datetime": "2023-10-01T00:00:00Z"},
                )

        .. tab-item:: Run Every 30 Minutes, 10 Times

            .. code-block:: python

                schedule = RecurringTimerSchedule(
                    interval_seconds=1800,
                    end={"condition": "iterations", "iterations": 10},
                )

        .. tab-item:: Run Every 10 Minutes, Indefinitely

            .. code-block:: python

                schedule = RecurringTimerSchedule(interval_seconds=600)

    Using these schedules, you can create a timer from a ``TransferData`` object:

    .. code-block:: pycon

        >>> from globus_sdk import TransferData, TransferTimer
        >>> schedule = ...
        >>> transfer_data = TransferData(...)
        >>> timer = TransferTimer(
        ...     name="my timer",
        ...     schedule=schedule,
        ...     body=transfer_data,
        ... )

    Submit the timer to the Timers service with
    :meth:`create_timer <globus_sdk.TimersClient.create_timer>`.
    """

    def __init__(
        self,
        *,
        name: str | MissingType = MISSING,
        schedule: dict[str, t.Any] | RecurringTimerSchedule | OnceTimerSchedule,
        body: dict[str, t.Any] | TransferData,
    ) -> None:
        super().__init__()
        self["timer_type"] = "transfer"
        self["name"] = name
        self["schedule"] = schedule
        self["body"] = self._preprocess_body(body)

    def _preprocess_body(
        self, body: dict[str, t.Any] | TransferData
    ) -> dict[str, t.Any]:
        # shallow-copy for dicts, convert any TransferData to a dict
        new_body = dict(body)
        # remove the skip_activation_check and submission_id parameters unconditionally
        # (not supported in timers, but often present in TransferData)
        new_body.pop("submission_id", None)
        new_body.pop("skip_activation_check", None)
        return new_body


class FlowTimer(GlobusPayload):
    """
    A helper for defining a payload for Flow Timer creation.
    Use this along with :meth:`create_timer <globus_sdk.TimersClient.create_timer>` to
    create a timer.

    ..  note::

        ``TimersClient`` has two methods for creating timers:
        ``create_timer`` and ``create_job``.

        This helper class only works with the ``create_timer`` method.

    :param flow_id: The flow ID to run when the timer runs.
    :param name: A name to identify this timer.
    :param schedule: The schedule on which the timer runs
    :param body: A transfer payload for the timer to use.

    The ``schedule`` field determines when the timer will run.
    Timers may be "run once" or "recurring", and "recurring" timers may specify an end
    date or the number of executions after which the timer will stop.
    A ``schedule`` is specified as a dict, but the SDK provides two helpers
    for constructing these data.

    **Example Schedules**

    .. tab-set::

        .. tab-item:: Run Once, Right Now

            .. code-block:: python

                schedule = OnceTimerSchedule()

        .. tab-item:: Run Once, At a Specific Time

            .. code-block:: python

                schedule = OnceTimerSchedule(datetime="2023-09-22T00:00:00Z")

        .. tab-item:: Run Every 5 Minutes, Until a Specific Time

            .. code-block:: python

                schedule = RecurringTimerSchedule(
                    interval_seconds=300,
                    end={"condition": "time", "datetime": "2023-10-01T00:00:00Z"},
                )

        .. tab-item:: Run Every 30 Minutes, 10 Times

            .. code-block:: python

                schedule = RecurringTimerSchedule(
                    interval_seconds=1800,
                    end={"condition": "iterations", "iterations": 10},
                )

        .. tab-item:: Run Every 10 Minutes, Indefinitely

            .. code-block:: python

                schedule = RecurringTimerSchedule(interval_seconds=600)

    Using these schedules, you can create a timer:

    .. code-block:: pycon

        >>> from globus_sdk import FlowTimer
        >>> schedule = ...
        >>> timer = FlowTimer(
        ...     name="my timer",
        ...     flow_id="00000000-19a9-44e6-9c1a-867da59d84ab",
        ...     schedule=schedule,
        ...     body={
        ...         "body": {
        ...             "input_key": "input_value",
        ...         },
        ...         "run_managers": [
        ...             "urn:globus:auth:identity:11111111-be6a-473a-a027-4cfe4ceeafe3"
        ...         ],
        ...     },
        ... )

    Submit the timer to the Timers service with
    :meth:`create_timer <globus_sdk.TimersClient.create_timer>`.
    """

    def __init__(
        self,
        *,
        flow_id: uuid.UUID | str,
        name: str | MissingType = MISSING,
        schedule: dict[str, t.Any] | RecurringTimerSchedule | OnceTimerSchedule,
        body: dict[str, t.Any],
    ) -> None:
        super().__init__()
        self["timer_type"] = "flow"
        self["flow_id"] = flow_id
        self["name"] = name
        self["schedule"] = schedule
        self["body"] = self._preprocess_body(body)

    def _preprocess_body(self, body: dict[str, t.Any]) -> dict[str, t.Any]:
        # Additional processing may be added in the future.
        return body.copy()


class RecurringTimerSchedule(GlobusPayload):
    """
    A helper used as part of a *timer* to define when the *timer* will run.

    A ``RecurringTimerSchedule`` is used to describe a *timer* which runs repeatedly
    until some end condition is reached.

    :param interval_seconds: The number of seconds between each run of the timer.
    :param start: The time at which to start the timer, either as an ISO 8601 string
        with timezone information, or as a ``datetime.datetime`` object.
    :param end: The end condition for the timer, as a dict. This either expresses a
        number of iterations for the timer or an end date.

    Example ``end`` conditions:

    .. code-block:: python

        # run 10 times
        end = {"condition": "iterations", "iterations": 10}

        # run until a specific date
        end = {"condition": "time", "datetime": "2023-10-01T00:00:00Z"}

    If the end condition is ``time``, then the ``datetime`` value can be expressed as a
    python ``datetime`` type as well, e.g.

    .. code-block:: python

        # end in 10 days
        end = {
            "condition": "time",
            "datetime": datetime.datetime.now() + datetime.timedelta(days=10),
        }
    """

    def __init__(
        self,
        interval_seconds: int,
        start: str | dt.datetime | MissingType = MISSING,
        end: dict[str, t.Any] | MissingType = MISSING,
    ) -> None:
        super().__init__()
        self["type"] = "recurring"
        self["interval_seconds"] = interval_seconds
        self["start"] = _format_date(start)
        self["end"] = end

        # if a datetime is given for part of the end condition, format it (and
        # shallow-copy the end condition)
        # primarily, this handles
        #    end={"condition": "time", "datetime": <some-datetime>}
        if isinstance(end, dict):
            self["end"] = {
                k: (_format_date(v) if isinstance(v, dt.datetime) else v)
                for k, v in end.items()
            }


class OnceTimerSchedule(GlobusPayload):
    """
    A helper used as part of a *timer* to define when the *timer* will run.

    A ``OnceTimerSchedule`` is used to describe a *timer* which runs exactly once.
    It may be scheduled for a time in the future.

    :param datetime: The time at which to run the timer, either as an ISO 8601
        string with timezone information, or as a ``datetime.datetime`` object.
    """

    def __init__(
        self,
        datetime: str | dt.datetime | MissingType = MISSING,
    ) -> None:
        super().__init__()
        self["type"] = "once"
        self["datetime"] = _format_date(datetime)


class TimerJob(GlobusPayload):
    r"""
    .. warning::

        This method of specifying and creating Timers for data transfer is now
        deprecated. Users should use ``TimerData`` instead.

        ``TimerJob`` is still supported for non-transfer use-cases.

    Helper for creating a timer in the Timers service. Used as the ``data``
    argument in :meth:`create_job <globus_sdk.TimersClient.create_job>`.

    The ``callback_url`` parameter should always be the URL used to run an
    action provider.

    :param callback_url: URL for the action which the Timers job will use.
    :param callback_body: JSON data which Timers will send to the Action Provider on
        each invocation
    :param start: The datetime at which to start the Timers job.
    :param interval: The interval at which the Timers job should recur. Interpreted as
        seconds if specified as an integer. If ``stop_after_n == 1``, i.e. the job is
        set to run only a single time, then interval *must* be None.
    :param name: A (not necessarily unique) name to identify this job in Timers
    :param stop_after: A date after which the Timers job will stop running
    :param stop_after_n: A number of executions after which the Timers job will stop
    :param scope: Timers defaults to the Transfer 'all' scope. Use this parameter to
        change the scope used by Timers when calling the Transfer Action Provider.

    .. automethodlist:: globus_sdk.TimerJob
    """

    def __init__(
        self,
        callback_url: str,
        callback_body: dict[str, t.Any],
        start: dt.datetime | str,
        interval: dt.timedelta | int | None,
        *,
        name: str | None = None,
        stop_after: dt.datetime | None = None,
        stop_after_n: int | None = None,
        scope: str | None = None,
    ) -> None:
        super().__init__()
        self["callback_url"] = callback_url
        self["callback_body"] = callback_body
        if isinstance(start, dt.datetime):
            self["start"] = start.isoformat()
        else:
            self["start"] = start
        if isinstance(interval, dt.timedelta):
            self["interval"] = int(interval.total_seconds())
        else:
            self["interval"] = interval
        if name is not None:
            self["name"] = name
        if stop_after is not None:
            self["stop_after"] = stop_after.isoformat()
        if stop_after_n is not None:
            self["stop_after_n"] = stop_after_n
        if scope is not None:
            self["scope"] = scope


def _format_date(date: str | dt.datetime | MissingType) -> str | MissingType:
    if isinstance(date, dt.datetime):
        if date.tzinfo is None:
            date = date.astimezone(dt.timezone.utc)
        return date.isoformat(timespec="seconds")
    else:
        return date



from __future__ import annotations

import typing as t
import uuid

from globus_sdk import AuthLoginClient, ConfidentialAppAuthClient, GlobusSDKUsageError
from globus_sdk.gare import GlobusAuthorizationParameters
from globus_sdk.scopes import Scope

from .app import GlobusApp
from .authorizer_factory import ClientCredentialsAuthorizerFactory
from .config import DEFAULT_CONFIG, GlobusAppConfig


class ClientApp(GlobusApp):
    """
    A ``GlobusApp`` for managing authentication state of a service account for use
    in service clients.

    A ``ClientApp`` requires the use of a confidential client created in a
    `Globus Project <https://app.globus.org/settings/developers>`. Client info may
    be passed either with the **client_id** and **client_secret** parameters or as
    a full **login_client**.

    ``ClientApps`` are configured by supplying a :class:`GlobusAppConfig` object to the
    **config** parameter. Of note however, **login_flow_manager** must not be set; a
    ``ClientApp`` does not use a login flow manager.

    See :class:`GlobusApp` for method signatures.

    .. rubric:: Example Usage:

    .. code-block:: python

        app = ClientApp("myapp", client_id=CLIENT_ID, client_secret=CLIENT_SECRET)
        transfer_client = TransferClient(app=app)
        res = transfer_client.endpoint_search("Tutorial Collection")


    :param app_name: A human-readable string to identify this app.
    :param login_client: A login client bound to a specific native client id or
        confidential client id/secret. Mutually exclusive with **client_id** and
        **client_secret**.
    :param client_id: A confidential client ID. Mutually exclusive with
        **login_client**.
    :param client_secret: A confidential client secret. Mutually exclusive with
        **login_client**.
    :param scope_requirements: A mapping of resource server to initial scope
        requirements.
    :param config: A data class containing configuration parameters for the app.
    """

    _login_client: ConfidentialAppAuthClient
    _authorizer_factory: ClientCredentialsAuthorizerFactory  # type: ignore

    def __init__(
        self,
        app_name: str = "Unnamed Globus App",
        *,
        login_client: ConfidentialAppAuthClient | None = None,
        client_id: uuid.UUID | str | None = None,
        client_secret: str | None = None,
        scope_requirements: (
            t.Mapping[str, str | Scope | t.Iterable[str | Scope]] | None
        ) = None,
        config: GlobusAppConfig = DEFAULT_CONFIG,
    ) -> None:
        if config.login_flow_manager is not None:
            raise GlobusSDKUsageError("A ClientApp cannot use a login_flow_manager")

        if login_client and not isinstance(login_client, ConfidentialAppAuthClient):
            raise GlobusSDKUsageError(
                "A ClientApp must use a ConfidentialAppAuthClient for its login_client"
            )

        super().__init__(
            app_name,
            login_client=login_client,
            client_id=client_id,
            client_secret=client_secret,
            scope_requirements=scope_requirements,
            config=config,
        )

    def _initialize_login_client(
        self,
        app_name: str,
        config: GlobusAppConfig,
        client_id: uuid.UUID | str,
        client_secret: str | None,
    ) -> AuthLoginClient:
        if not client_secret:
            raise GlobusSDKUsageError(
                "A ClientApp requires a client_secret to initialize its own login "
                "client."
            )

        return ConfidentialAppAuthClient(
            client_id=client_id,
            client_secret=client_secret,
            app_name=app_name,
            environment=config.environment,
        )

    def _initialize_authorizer_factory(self) -> None:
        self._authorizer_factory = ClientCredentialsAuthorizerFactory(
            token_storage=self.token_storage,
            confidential_client=self._login_client,
            scope_requirements=self._scope_requirements,
        )

    def _run_login_flow(
        self, auth_params: GlobusAuthorizationParameters | None = None
    ) -> None:
        """
        Run an authorization flow to get new tokens which are stored and available
        for the next authorizer gotten by get_authorizer.

        As a ClientApp this is just a client credentials call.

        :param auth_params: A GlobusAuthorizationParameters to control authentication
            only the required_scopes parameter is used.
        """
        auth_params = self._auth_params_with_required_scopes(auth_params)
        if not auth_params.required_scopes:
            raise GlobusSDKUsageError(
                "A ClientApp cannot get tokens without configured required scopes."
            )
        token_response = self._login_client.oauth2_client_credentials_tokens(
            requested_scopes=auth_params.required_scopes
        )
        self._authorizer_factory.store_token_response_and_clear_cache(token_response)
